package leveldb

import (
	"bytes"
	"encoding/binary"
	"log"
	"net"
	"time"

	"github.com/golang/protobuf/proto"
	"github.com/syndtr/goleveldb/leveldb"
	"github.com/syndtr/goleveldb/leveldb/errors"
	"github.com/syndtr/goleveldb/leveldb/filter"
	"github.com/syndtr/goleveldb/leveldb/opt"
	"github.com/syndtr/goleveldb/leveldb/util"

	ippb "git.autistici.org/ai3/tools/iprep/proto"
)

var (
	eventPrefix     = []byte("e/")
	ipIndexPrefix   = []byte("i/ip/")
	typeIndexPrefix = []byte("i/t/")
)

type DB struct {
	db *leveldb.DB
}

func Open(path string) (*DB, error) {
	opts := &opt.Options{
		BlockCacheCapacity: 64 * opt.MiB,
		WriteBuffer:        64 * opt.MiB,
		Filter:             filter.NewBloomFilter(12),
	}
	db, err := leveldb.OpenFile(path, opts)
	if errors.IsCorrupted(err) {
		log.Printf("database appears to be corrupted: %v, attempting recovery...", err)
		db, err = leveldb.RecoverFile(path, opts)
	}
	if err != nil {
		return nil, err
	}
	return &DB{db: db}, nil
}

func (db *DB) Close() {
	db.db.Close()
}

func (db *DB) AddAggregate(aggr *ippb.Aggregate) error {
	var ev ippb.Event

	wb := new(leveldb.Batch)
	for _, bt := range aggr.ByType {
		for _, item := range bt.ByIp {
			id := nextID()

			ev.Type = bt.Type
			ev.Ip = item.Ip
			ev.Count = item.Count
			data, err := proto.Marshal(&ev)
			if err != nil {
				return err
			}

			wb.Put(eventKeyFromID(id), data)
			wb.Put(indexKeyFromID(id, ipIndexPrefix, ipToBytes(ev.Ip)), data)
			wb.Put(indexKeyFromID(id, typeIndexPrefix, []byte(ev.Type)), data)
		}
	}

	return db.db.Write(wb, nil)
}

func (db *DB) delete(endTime time.Time) error {
	iter := db.db.NewIterator(prefixUntil(endTime, eventPrefix), &opt.ReadOptions{
		DontFillCache: true,
	})

	wb := new(leveldb.Batch)
	for iter.Next() {
		var e ippb.Event
		if err := proto.Unmarshal(iter.Value(), &e); err != nil {
			continue
		}
		id := eventIDFromKey(iter.Key())
		wb.Delete(iter.Key())
		wb.Delete(indexKeyFromID(id, ipIndexPrefix, ipToBytes(e.Ip)))
		wb.Delete(indexKeyFromID(id, typeIndexPrefix, []byte(e.Type)))
	}
	return db.db.Write(wb, nil)
}

func (db *DB) WipeOldData(age time.Duration) error {
	endTime := time.Now().Add(-age)
	db.delete(endTime) // nolint
	return db.db.CompactRange(util.Range{})
}

func (db *DB) ScanIP(startTime time.Time, ip string) (map[string]int64, error) {
	iter := db.db.NewIterator(prefixSince(startTime, ipIndexPrefix, ipToBytes(ip)), &opt.ReadOptions{
		DontFillCache: true,
	})
	m := make(map[string]int64)
	for iter.Next() {
		var err error
		var e ippb.Event
		if err = proto.Unmarshal(iter.Value(), &e); err != nil {
			continue
		}
		m[e.Type] += e.Count
	}
	return m, iter.Error()
}

func (db *DB) ScanType(startTime time.Time, t string) (map[string]int64, error) {
	iter := db.db.NewIterator(prefixSince(startTime, typeIndexPrefix, []byte(t)), &opt.ReadOptions{
		DontFillCache: true,
	})
	m := make(map[string]int64)
	for iter.Next() {
		var err error
		var e ippb.Event
		if err = proto.Unmarshal(iter.Value(), &e); err != nil {
			continue
		}
		m[e.Ip] += e.Count
	}
	return m, iter.Error()
}

func (db *DB) Compact() error {
	return db.db.CompactRange(util.Range{})
}

func eventKeyFromID(id []byte) []byte {
	return append(eventPrefix, id...)
}

func eventIDFromKey(key []byte) []byte {
	return key[len(eventPrefix):]
}

func indexKeyFromID(id, indexPrefix, indexKey []byte) []byte {
	return bytes.Join([][]byte{indexPrefix, indexKey, id}, []byte{})
}

func ipToBytes(s string) []byte {
	return []byte(net.ParseIP(s))
}

func prefixSince(start time.Time, indexPrefix, indexKey []byte) *util.Range {
	pfx := append(indexPrefix, indexKey...)
	b := make([]byte, len(pfx)+8)
	copy(b[:len(pfx)], pfx)
	binary.BigEndian.PutUint64(b[len(pfx):len(pfx)+8], uint64(start.UnixNano()))
	pfx[len(pfx)-1]++
	return &util.Range{
		Start: b,
		Limit: pfx,
	}
}

func prefixUntil(end time.Time, pfx []byte) *util.Range {
	b := make([]byte, len(pfx)+8)
	copy(b[:len(pfx)], pfx)
	binary.BigEndian.PutUint64(b[len(pfx):len(pfx)+8], uint64(end.UnixNano()))
	return &util.Range{
		Limit: b,
	}
}
